import sys, argparse
from .config import Config

from .cmd_use import use
from .cmd_project import project, projects
from .cmd_get import get
from .cmd_describe import describe
from .cmd_log import log
from .cmd_whoami import whoami

def main():
    # Common parser, with shared arguments for all subcommands:
    common = argparse.ArgumentParser(add_help=False)
    common.add_argument("-n", "--namespace", dest="namespace")
    common.add_argument("-A", "--all-namespaces", dest="all_namespaces",action='store_true')

    # Main Parser for sub commands
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers()

    # omg use </path/to/must-gather>
    p_use = subparsers.add_parser('use', parents=[common], aliases=['login'],
                                  help='Select the must-gather to use')
    p_use.add_argument('mg_path', metavar='must-gather-location', type=str)
    p_use.set_defaults(func=use)

    # omg project
    p_project = subparsers.add_parser('project', parents=[common],
                                      help='Display information about the current active project and existing projects')
    p_project.set_defaults(func=project)

    # omg projects
    p_projects = subparsers.add_parser('projects', parents=[common],
                                       help='Display information about the current active project and existing projects')
    p_projects.set_defaults(func=projects)

    # omg get <object(s)>
    p_get = subparsers.add_parser('get', parents=[common],
                                  help='Display one or many resources')
    p_get.add_argument('objects', nargs='*', type=str)
    p_get.add_argument("-o", "--output", dest="output",
                       choices=['yaml', 'json', 'wide'] )
    p_get.set_defaults(func=get)

    # omg describe <object(s)>
    p_describe = subparsers.add_parser('describe', parents=[common],
                                       help='This command joins many API calls together to form a detailed description of a given resource.')
    p_describe.add_argument('object', nargs='*', type=str)
    p_describe.set_defaults(func=describe)

    # omg log <pod>
    p_log = subparsers.add_parser('log', parents=[common],
                                  help='Display one or many resources')
    p_log.set_defaults(func=log)

    # omg whoami
    p_whoami = subparsers.add_parser('whoami', parents=[common],
                                  help='Display who you are')
    p_whoami.set_defaults(func=whoami)

    # process args and call the corresponding function
    args = parser.parse_args()
    try:
        func = args.func
    except AttributeError:
        parser.error("too few arguments")
    func(args)



import sys, os
from dateutil.parser import parse
from dateutil.relativedelta import relativedelta
from datetime import datetime, timedelta

from .config import Config

# Check if namespace(s) exists and
# returns their path(s) in must-gather
# Returns all if '_all' is passed
def locate_ns(ns):
    ns_dir = Config().path + '/namespaces/'
    if ns is None:
        print("[ERROR] Namespace not set. Select a project (omg project) or specify a namespace (-n)")
        sys.exit(1)
    elif ns != '_all':
        ns_path = [ ns_dir + ns ]
        if not os.path.isdir(ns_path[0]):
            print("[ERROR] Namespace not found: ", ns_path[0])
            sys.exit(1)
        return ns_path
    elif ns == '_all':
        ns_path = [ ns_dir+n for n in os.listdir(ns_dir) if os.path.isfile(ns_dir+n+'/core/pods.yaml')]
        if len(ns_path) > 0:
            return ns_path
        else:
            print("[ERROR] No namespace found")
            sys.exit(1)
    else:
        print('[BUG] This should not happen, please report')


# This function is used to calculate age of the objects
# We compare the time reported in the yaml definition,
# with the timestamp of the yaml file (i.e, when it was generated).
# This should give us age of the object at the time of must-gather
# obj_time time is in standard format e.g: '2020-06-04T22:10:41Z'
# file_ts is unix ts e.g: 1590912494.0 (returned by os.path.getmtime)

def age(obj_time, file_ts):

    try:
        dt1 = parse(obj_time, ignoretz=True)
        dt2 = datetime.fromtimestamp(file_ts)
        rd = relativedelta(dt2, dt1)
    except:
        return '?'

    if rd.days > 0:
        return str(rd.days)+'d'
    elif rd.hours > 9:
        return str(rd.hours)+'h'
    elif rd.hours > 0 and rd.hours < 10:
        return str(rd.hours)+'h'+str(rd.minutes)+'m'
    elif rd.minutes > 9:
        return str(rd.minutes)
    elif rd.minutes > 0 and rd.minutes < 10:
        return str(rd.minutes)+'m'+str(rd.seconds)+'s'
    else:
        return str(rd.seconds)+'s'


import sys, os, yaml, json, tabulate
from tabulate import tabulate
from .config import Config
from .resource_map import check_res
from .helper import locate_ns
from .helper import age

# This function gets the simple namespace bound resource types like:
# pod, event, configmap, secret, etc.
def simple(ns, yaml_loc, names):
    # We need ns for this resource type
    if ns is None:
        print('[ERROR] No namespaces selected. Use `omg project` or specify -n')
        sys.exit(1)
    ns_path = locate_ns(ns)
    # Collect the resources
    collected=[]
    for n in ns_path:
        yaml_file = n + yaml_loc
        # record when was this yaml generated (to calc age)
        gen_ts = os.path.getmtime(yaml_file)
        try:
            with open(yaml_file, 'rb') as yf:
                try:
                    res = yaml.safe_load(yf)
                except:
                    print("[ERROR] Parsing error in ", yaml_file)
                    sys.exit(1)
        except:
            print("[ERROR] Could not read file:", yaml_file)
            sys.exit(1)
        collected.extend([ {'res':r,'gen_ts':gen_ts} for r in res['items'] if r['metadata']['name'] in names or '_all' in names ])
    return collected

def simple_out(res, ns, output):
    output_res=[[]]
    # header
    if ns == '_all':
        output_res[0].append('NAMESPACE')
    output_res[0].extend(['NAME','AGE'])
    # resources
    for r in res:
        p = r['res']
        row = []
        # namespace (for --all-namespaces)
        if ns == '_all':
            row.append(p['metadata']['namespace'])
        # name
        row.append(p['metadata']['name'])
        # age
        try:
            ct = p['metadata']['creationTimestamp']
            ts = r['gen_ts']
            row.append(age(ct,ts))
        except:
            row.append('Unknown')

        output_res.append(row)

    print(tabulate(output_res,tablefmt="plain"))


def pod_out(res, ns, output):
    # Generate output table if -o not set or 'wide'
    # We will create an array of array and then print if with tabulate
    if output in [None,'wide']:
        output_pods=[[]]
        # header
        if ns == '_all':
            output_pods[0].append('NAMESPACE')
        if output == 'wide':
            output_pods[0].extend(['NAME','READY','STATUS','RESTARTS','AGE','IP','NODE'])
        else:
            output_pods[0].extend(['NAME','READY','STATUS','RESTARTS','AGE'])
        # pods
        for pod in res:
            p = pod['res']
            row = []
            # namespace (for --all-namespaces)
            if ns == '_all':
                row.append(p['metadata']['namespace'])
            # name
            row.append(p['metadata']['name'])
            # containers/ready count
            containers = str(len(p['spec']['containers']))
            containers_ready = str(len([ r for r in p['status']['containerStatuses']  if r['ready'] == True ]))
            row.append(containers_ready+'/'+containers)
            # status
            row.append(p['status']['phase'])
            # restarts
            row.append(max([ r['restartCount'] for r in p['status']['containerStatuses'] ]))
            # age
            pod_ct = p['metadata']['creationTimestamp']
            gen_ts = pod['gen_ts']
            row.append(age(pod_ct,gen_ts))
            # pod ip and node (if -o wide)
            if output == "wide":
                row.append(p['status']['podIP'])
                row.append(p['spec']['nodeName'])

            output_pods.append(row)

        print(tabulate(output_pods,tablefmt="plain"))



# The high level function that gets called for any "get" command
# Checks what objects to get,
# and call the respective omg_get_<obj> function
def get(a):
    # Check if -A/--all-namespaces is set
    # else, Set the namespace
    # -n/--namespace takes precedence over current project 
    if a.all_namespaces is True:
        ns = '_all'
    else:
        if a.namespace is not None:
            ns = a.namespace
        elif Config().project is not None:
            ns = Config().project
        else:
            ns = None

    # Process the objects, they can be in various formats e.g:
    #   pod httpd, dc/httpd, routes , svc/httpd, pod,svc etc.
    # We get all these args (space separated) in the array args.objects
    # We'll process them and normalize them in a python dict
    objects = {}

    last_object = []
    for o in a.objects:
        # Case where we have a '/'
        # e.g omg get pod/httpd
        if '/' in o:
            o_type = check_res(o.split('/')[0])
            o_name = o.split('/')[1]
            if o_type is not None:
                if o_type['type'] in objects:
                    objects[o_type['type']].append(o_name)
                else:
                    objects[o_type['type']] = [o_name]
            else:
                print("[ERROR] Invalid object type: ",o.split('/')[0])
                sys.exit(1)
        # Case where we have a ','
        # e.g omg get dc,svc httpd
        elif ',' in o:
            if not last_object:
                o_types = o.split(',')
                for ot in o_types:
                    if check_res(ot) is None:
                        print("[ERROR] Invalid object type: ",ot)
                        sys.exit(1)
                last_object = [ check_res(x)['type'] for x in o_types]
            else:
                # last_object was set, meaning this should be object name
                print("[ERROR] Invalid resources to get: ", a.objects)
                sys.exit(1)
        # Simple word that is an object type
        # meaning previous word should not have been an object type
        # i.e, last_object should be None
        elif check_res(o) is not None and not last_object:
            last_object = [ o ]
        # Simple word that is not an object type
        # meaning previous word should have been an object type
        # i.e, last_object should not be None
        elif check_res(o) is None and last_object:
            for lo in last_object:
                o_type = check_res(lo)
                if o_type['type'] in objects:
                    objects[o_type['type']].append(o)
                else:
                    objects[o_type['type']] = [o]
            last_object = []
        else:
            print("[ERROR] Invalid resources to get: ", o)
            sys.exit(1)
    # If after going through all the args, we have last_object set
    # We should process that (request of object type without name)
    # e.g, omg get pods or omg get pods,dc,svc etc.
    if last_object:
        for lo in last_object:
            o_type = check_res(lo)
            objects[o_type['type']] = ['_all']

    # Object based routing
    # i.e, call the function specific to the object type
    for o in objects.keys():
        get_func = check_res(o)['get_func']
        getout_func = check_res(o)['getout_func']
        yaml_loc = check_res(o)['yaml_loc']
        output = a.output

        res = get_func(ns, yaml_loc, objects[o])

        # Error out if no objects/resources were collected
        if len(res) == 0:
            print("No resources found.")
            sys.exit(1)
        # Genereate yaml if -o yaml
        elif output == 'yaml':
            if len(res) == 1:
                print(yaml.dump(res[0]['res']))
            elif len(res) > 1:
                print(yaml.dump({'apiVersion':'v1','items':[cp['res']for cp in res]}))
        # Generate json if -o json
        elif output == 'json':
            if len(res) == 1:
                print(json.dumps(res[0]['res'],indent=4))
            elif len(res) > 1:
                print(json.dumps({'apiVersion':'v1','items':[cp['res']for cp in res]},indent=4))
        # output using the getout_func
        else:
            getout_func(res, ns, output)


import sys, os
from .config import Config

def use(a):
    path = a.mg_path
    c = Config()
    if os.path.isdir(path) and os.path.isfile(path+'/version'):
        c.save(path=os.path.abspath(path))
    elif os.path.isdir(path) and os.path.isfile(path + '/' + str(os.listdir(path)[0]) + '/version'):
        c.save(path=os.path.abspath(path + '/' + str(os.listdir(path)[0])))
    else:
        print("[ERROR] Invalid must-gather path. Please point to the extracted must-gather directory")
        sys.exit(1)


import sys

from . import cmd_get as get

# This function normalizes/standarizes the different names of object/resources types
# Also returns the function that handles get/describe of this object type
# This is how the top level omg_get function knows which objects we can handle and how
# e.g: for the input 'pods', 'pod' or 'po' it will return: 
#   { 'type':'pod', 'get_func': omg_get_pod, 'desc_func': omg_desc_pod }
# returns None if not an object type that we can handle
def check_res(t):
    map = { 'pod': { 'aliases': ['pod', 'pods', 'po'],
                     'get_func': get.simple, 'getout_func': get.pod_out,
                     'yaml_loc': '/core/pods.yaml'},

            'svc': { 'aliases': ['service', 'svc'],
                     'get_func': get.simple, 'getout_func': get.simple_out,
                     'yaml_loc': '/core/services.yaml'},

            'secrets' : { 'aliases': ['secrets'],
                    'get_func': get.simple, 'getout_func': get.simple_out,
                    'yaml_loc': '/core/secrets.yaml'},

            'cm' : { 'aliases': ['configmap', 'configmaps', 'cm'],
                    'get_func': get.simple, 'getout_func': get.simple_out,
                    'yaml_loc': '/core/configmaps.yaml'},

            'ep' : { 'aliases': ['endpoint', 'endpoints', 'ep'],
                    'get_func': get.simple, 'getout_func': get.simple_out,
                    'yaml_loc': '/core/endpoints.yaml'},

            'ev' : { 'aliases': ['event', 'events', 'ev'],
                    'get_func': get.simple, 'getout_func': get.simple_out,
                    'yaml_loc': '/core/events.yaml'},

            'pvc' : { 'aliases': ['persistentvolumeclaim', 'persistentvolumeclaims', 'pvc'],
                    'get_func': get.simple, 'getout_func': get.simple_out,
                    'yaml_loc': '/core/persistentvolumeclaims.yaml'},

            'rc' : { 'aliases': ['replicationcontrollers', 'replicationcontrollers', 'rc'],
                    'get_func': get.simple, 'getout_func': get.simple_out,
                    'yaml_loc': '/core/replicationcontrollers.yaml'},
            }

    #return next(
    #    ( { 'type':x,
    #        'get_func': map[x]['get_func'],
    #        'getout_func': map[x]['getout_func'],
    #        'yaml_loc': map[x]['yaml_loc'] }
    #        for x in map.keys() if t.lower() in map[x]['aliases'] 
    #    ),None)
    for x in map.keys():
        if t.lower() in map[x]['aliases']:
            return { 'type':x,
                     'get_func': map[x]['get_func'],
                     'getout_func': map[x]['getout_func'],
                     'yaml_loc': map[x]['yaml_loc'] }
    return None


import sys, os, yaml

CONFIG_FILE = os.getenv("HOME") + "/.omgconfig"

class Config:

    path = None
    project = None

    def __init__(self):
        if os.path.exists(CONFIG_FILE):
            with open(CONFIG_FILE, 'r') as cf:
                c = yaml.safe_load(cf)
                if 'path' in c:
                    Config.path = c['path']
                if 'project' in c:
                    Config.project = c['project']

    def save(self, path=None, project=None):
        c = {}
        if path is not None:
            Config.path = path
            c['path'] = path
        if project is not None:
            Config.project = project
            c['project'] = project
        try:
            with open(CONFIG_FILE, 'w') as cf:
                yaml.dump(c, cf, default_flow_style=False)
        except IOError:
            print("[ERROR] Could not write config file:", CONFIG_FILE)
